# <path to game directory>/addons/eventscripts/_engines/python/Lib/esutils/
# delays.py
# by Adam Cunnington

"""Provide optimised methods which enable a user to delay the execution of 
callables in the primary gamethread.

"""

import bisect
import functools
import time

import es

__all__ = (
    "DelaysError", 
    "Delay", 
)

class DelaysError(Exception):
    """General error encountered relating to admin.libraries.delays"""
    
class _DelayManager(object):
    def __init__(self):
        self.delays = []
        
    def _tick(self):
        if not self.delays:
            return
        call_time, delay = self.delays.pop()
        if time.time() < call_time:
            return
        delay.callable_name()
        if not delay.recurring:
            delay.running = False
            if not self.delays:
                es.addons.unregisterTickListener(_TICK_LISTENER)
            return
        else:
            delay.call_time = time.time() + delay.interval
            bisect.insort_left(self.delays, (delay.call_time, delay))
            
_delay_manager = _DelayManager()

_TICK_LISTENER = _delay_manager._tick

class Delay(object):
    """Recurringly or singularly call a given name every <interval> seconds."""
    
    def __init__(self, callable_name, *args, **kwargs):
        """Instantiate a Delay object.
        
        Arguments:
        callable_name - the name of the callable.
        args - the positional arguments to be passed to the callable.
        kwargs - the keyword arguments to be passed to the callable.
        
        """
        self.callable_name = functools.partial(callable_name, *args, **kwargs)
        self.running = False
        
    def start(self, interval, recurring=False, call_now=False):
        """Start the delay.
        
        Arguments:
        interval - the number of seconds to delay the call by.
        recurring (Keyword Default: False) - whether or not the delay should 
        happen recurringly or just be a singular occurence.
        call_now (Keyword Default: False) - whether or not the delayed 
        callable should also be called now.
        
        """
        if self.running:
            raise DelaysError("this delay object is already running")
        self.interval = interval
        self.recurring = recurring
        self.call_time = time.time() + interval
        if not _delay_manager.delays:
            es.addons.registerTickListener(_TICK_LISTENER)
        bisect.insort_left(_delay_manager.delays, (self.call_time, self))
        self.running = True
        if call_now:
            self.callable_name()
            
    def stop(self):
        """"Cancel the delay and break any looped call."""
        if not self.running:
            raise DelaysError("this delay object is not running")
        _delay_manager.delays.remove((self.call_time, self))
        self.running = False
        if not _delay_manager.delays:
            es.addons.unregisterTickListener(_TICK_LISTENER)