# <path to game directory>/addons/eventscripts/_engines/python/Lib/esutils/
# menus/__init__.py
# by Adam Cunnington

"""Provide a powerful API to allow the user to create & manage menus 
consisting of options. Also, manage menu selections and send them to an 
explictly registered callback function.

"""

from esutils import admins, delays
import es

__all__ = (
    "MenusError", 
    "menu_manager", 
    "Menu", 
    "DynamicMenu", 
    "MenuItem", 
    "MenuOption", 
    "close_popup", 
    "get_popup", 
)

class MenusError(Exception):
    """Generic error encountered relating to admin.libraries.menus"""
    
class _Player(object):
    def __init__(self):
        self._delay = None
        self.history = []
        self.identifiers = []
        self.popup = None
        
class _MenuManager(object):
    # Encase names that would otherwise be global in order to provide a basis 
    # with the quality of developability.
    
    def __init__(self):
        self.players = {}
        
# Public variable as in certain circumstances, users may wish to access player 
# information.
menu_manager = _MenuManager()

class Menu(object):
    """Create a dynamically constructed menu that can be sent and 
    simultaneously viewed by multiple users.
    
    """

    def __init__(self, callback=None):
        """Instantiate a Menu object.
        
        Arguments:
        callback (Keyword Default: None) - the name to call when a menu 
        item is selected.
        
        """
        self.callback = callback
        self._title = " "
        self._description = " "
        self.page_format = "[Page %(current_page)s/%(total_pages)s]"
        self._separator = "------------------------------"
        self.items = []
        self._max_items = 6
        
    def _build(self, user_id, page):
        total_items = len(self.items)
        total_pages, remainder = divmod(total_items, self._max_items)
        if remainder:
            total_pages += 1
        selectable_keys = ""
        lines = []
        lines.append(self._title)
        lines.append(self._description)
        lines.append(self.page_format % {"current_page": page, 
                                         "total_pages": total_pages})
        lines.append(self._separator)
        start = (page - 1) * self._max_items
        for i in xrange(self._max_items):
            index = start + i
            number = i + 1
            if index < total_items:
                item = self.items[index]
                if item.selectable:
                    steam_id = es.getplayersteamid(user_id)
                    if (not item.requires_auth or 
                        admins.is_user_authorised(steam_id, item.permission)):
                        lines.append("->%s. %s" %(number, item.text))
                        selectable_keys += str(number)
                    else:
                        lines.append("%s. %s" %(number, item.text))
                else:
                    lines.append(item.text)
            else:
                lines.extend((" ", ) * (self._max_items - (number - 1)))
                break
        lines.append(self._separator)
        if menu_manager.players[user_id].history:
            number = "7"
            lines.append("->%s. Previous Popup" % number)
            selectable_keys += number
        else:
            # In order to make the sizes of menus uniform so that navigation 
            # between menus looks professional, a blank line needs to be 
            # added. Apparently, a space is the minimum for a blank line.
            lines.append(" ")
        if page > 1:
            # This will occur if the user has been explicitly sent a page from 
            # a menu greater than 1 with no previous history. In this 
            # instance, we want them to be able to access previous pages in 
            # the menu.
            number = "8"
            lines.append("->%s. Previous Page" % number)
            selectable_keys += number
        else:
            lines.append(" ")
        if page < total_pages:
            number = "9"
            lines.append("->%s. Next Page" % number)
            selectable_keys += number
        else:
            lines.append(" ")
        number = "0"
        lines.append("%s. Close" % number)
        selectable_keys += number
        return "\n".join(lines), selectable_keys
        
    def _get_description(self):
        return self._description
        
    def _get_max_items(self):
        return self._max_items
        
    def _get_separator(self):
        return self._separator
        
    def _get_title(self):
        return self._title
        
    def _set_description(self, value):
        _validate_true_value(value)
        self._description = value
        
    description = property(_get_description, _set_description)
    
    def _set_max_items(self, value):
        if not 0 < value < 7:
            raise ValueError("expected 0 < value < 7")
        self._max_items = value
        
    max_items = property(_get_max_items, _set_max_items)
    
    def _set_separator(self, value):
        _validate_true_value(value)
        self._separator = value
        
    separator = property(_get_separator, _set_separator)
    
    def _set_title(self, value):
        _validate_true_value(value)
        self._title = value
        
    title = property(_get_title, _set_title)
    
    def send(self, user_id, page=1, reset_navigation=True):
        """Build the popup afresh and send it to the user.
        
        Arguments:
        user_id - the unique session ID of the user who should receive the 
        popup.
        page (Keyword Default: 1) - the page of the menu that should be sent 
        to the user.
        reset_navigation (Keyword Default: True) - whether or not the player's 
        history should be forgotten.
        
        """
        if reset_navigation:
            player = menu_manager.players[user_id] = _Player()
        else:
            player = menu_manager.players[user_id]
        es.menu(0, user_id, *self._build(user_id, page))
        # As a result of valve's initiatve, a value of 0 passed to the timeout 
        # parameter no longer means infinite. The menu will automatically 
        # disappear after about 5 seconds so it needs to be constantly resent.
        player._delay = delays.LoopingCall(self.send, user_id, page, 
                                           reset_navigation)
        player._delay.start(4.5, False, False)
        player.popup = self, page
        
class DynamicMenu(Menu):
    """Create a dynamically constructed menu that can be sent and 
    simultaneously viewed by multiple users. Menu options are dynamically 
    fetched instead of being statically added.
    
    """
    
    def __init__(self, item_getter, callback=None):
        """Instantiate a DynamicMenu object.
        
        Arguments:
        item_getter - the name to call to dynamically fetch the menu's options.
        callback (Keyword Default: None) - the name to call when a menu item 
        is selected.
        
        """
        self.item_getter = item_getter
        super(DynamicMenu, self).__init__(callback)
        
    def send(self, *args, **kwargs):
        """Build the popup afresh with dynamic options and send it to the 
        user.
        
        Arguments:
        args - the positional arguments defined by the parent class.
        kwargs - the keyword arguments defined by the parent class.
        
        """
        self.items = self.item_getter()
        super(DynamicMenu, self).send(*args, **kwargs)
        
class MenuItem(object):
    """Create a text-only menu item that cannot be selected."""
    
    def __init__(self, text):
        """Instantiate a MenuItem object.
        
        Arguments:
        text - the text to appear in the menu for this item.
        
        """
        self.text = text
        self.selectable = False
        
class MenuOption(object):
    """Create an numbered menu item that can be selected."""

    def __init__(self, text, identifier=None, selectable=True):
        """Instantiate a MenuOption object.
        
        Arguments:
        args - the positional arguments defined by the parent class.
        identifier (Keyword Default: None) - the value to be passed to the 
        menu's callback when selected.
        selectable (Keyword Default: True) - whether or not the option can be
        selected.
        kwargs - the keyword arguments defined by the parent class.
        
        """
        self.text = text
        self.identifier = identifier
        self.selectable = selectable
        self.requires_auth = False
        self.subpopup = None
        
    def set_permission(self, permission=None):
        """Cause authorisation to be required by the user to select the 
        option.
        
        Arguments:
        permission (Keyword Default: None) - the name of the permission 
        requried.
        
        """
        self.requires_auth = True
        self.permission = permission
        
    def set_subpopup(self, menu, page=1, send_callback=False):
        """Cause a separate popup to be automatically sent to the user when 
        they select this option.
        
        Arguments:
        menu - the menu object to send.
        page (Keyword Default: 1) - the page of the menu to send.
        send_callback (Keyword Default: False) - whether or not the menu's 
        callback should still be called.
        
        """
        self.subpopup = menu, page
        self.send_callback = send_callback
        
def _validate_true_value(value):
    # Empty lines in a popup are ignored and so a space is required for a 
    # seemingly blank line.
    if not value:
        raise ValueError("expected a True value")
        
def _client_command_filter(user_id, args):
    if (args.pop(0).lower() != "menuselect" or not args or 
        not args[0].isdigit() or user_id not in menu_manager.players):
        # If the command submitted wasn't a valid menu select, we should exit 
        # the function immediately.
        return True
    selection = int(args[0])
    if selection not in xrange(1, 11):
        return True
    player = menu_manager.players[user_id]
    if player.popup is None:
        return True
    player._delay.stop()
    menu, page = player.popup
    player.history.append((menu, page))
    player.popup = None
    if selection == 7:
        del player.history[-1]
        previous_menu, previous_page = player.history.pop()
        previous_menu.send(user_id, previous_page, False)
    elif selection == 8:
        menu.send(user_id, page - 1, False)
    elif selection == 9:
        menu.send(user_id, page + 1, False)
    elif selection == 10:
        pass
    else:
        option = menu.items[((page - 1) * 7) + selection - 1]
        player.identifiers.append(option.identifier)
        if option.subpopup is not None:
            next_menu, next_page = option.subpopup
            next_menu.send(user_id, next_page, False)
            if not option.send_callback:
                return True
        menu.callback(user_id, player.identifiers)
    return True
    
es.addons.registerClientCommandFilter(_client_command_filter)

def _player_disconnect(event_var):
    user_id = int(event_var["userid"])
    if user_id in menu_manager.players:
        player = menu_manager.players[user_id]
        if player._delay is not None:
            player._delay.stop()
        del menu_manager.players[user_id]
        
es.addons.registerForEvent(menu_manager, "player_disconnect", 
                           _player_disconnect)
                           
def close_popup(user_id):
    """Effectively close a user's popup by cancelling the next resend. raise 
    an exception if the user does not have a menu open.
    
    Arguments:
    user_id - the unique session ID of the user who's menu should be closed.
    
    """
    if user_id in menu_manager.players:
        player = menu_manager.players[user_id]
        if player._delay is not None:
            player._delay.stop()
            return
    raise MenusError("user ID, %s has no menu active" % user_id)
    
def get_popup(user_id):
    """Return a 2-part tuple (menu, page) of the popup being currently viewed 
    by the user. Otherwise, return None.
    
    Arguments:
    user_id - the unique session ID of the user.
    
    """
    if user_id in menu_manager.players:
        return menu_manager.players[user_id].popup
    return None